"""Mail `download` command module."""

import errno
import os
from collections import defaultdict

from loguru import logger

from gmail.commands import GoogleMailCmd


class MailDownloadCmd(GoogleMailCmd):
    """Mail `download` command class."""

    def init_command(self) -> None:
        """Initialize mail `download` command."""

        parser = self.add_subcommand_parser(
            "download",
            help="download mail messages",
            description=self.cli.dedent(
                """
    The `%(prog)s` program downloads a mail message.
                """
            ),
        )

        parser.add_argument(
            "MSG_ID",
            help="the id of the message to download",
        )

    def run(self) -> None:
        """Run mail `download` command."""

        msg_id = self.options.MSG_ID

        unknown_mimetypes: dict[str, int] = defaultdict(int)

        for mimetype, filename, attachment_id in self.cli.api.get_next_attachment_id(msg_id):

            logger.error(
                f"mimetype={mimetype!r} filename={filename!r} attachment_id={attachment_id!r}"
            )

            if mimetype[:5] not in ("image", "video") and mimetype != self.cli.api.mimetype_PDF:
                logger.debug("mimeType {!r} not image/video/pdf", mimetype)
                unknown_mimetypes[mimetype] += 1
                continue

            print("Downloading", filename)

            try:
                os.makedirs(self.cli.api.download_dir)
            except OSError as err:
                if err.errno != errno.EEXIST:
                    raise  # pragma: no cover

            with open(filename, "wb") as _:
                _.write(self.cli.api.get_attachment_data(msg_id, attachment_id))

        #
        for mimetype, count in unknown_mimetypes.items():
            print(str.format("unknown mimeType {:5d} {:s}", count, mimetype))



"""Mail `list` command module."""

from time import localtime, strftime
from typing import Any

from loguru import logger

from gmail.api import GoogleMailAPI
from gmail.commands import GoogleMailCmd


class MailListCmd(GoogleMailCmd):
    """Mail `list` command class."""

    def init_command(self) -> None:
        """Initialize mail `list` command."""

        parser = self.add_subcommand_parser(
            "list",
            help="list mail messages",
            description=self.cli.dedent(
                """
    The `%(prog)s` program lists mail messages.
                """
            ),
        )

        group = parser.add_argument_group(
            "Printing options",
            "These options are mutually exclusive.",
        )

        exc = group.add_mutually_exclusive_group()
        exc.add_argument(
            "--print-message",
            "--print-msg",
            action="store_true",
            help="print message",
        )
        exc.add_argument(
            "--print-listing",
            action="store_true",
            help="print listing",
        )
        self.add_pretty_print_option(exc)

        group = parser.add_argument_group("Filtering options")

        group.add_argument(
            "--msg-id",
            "--msgid",
            help="operate on `MSG_ID` only",
        )

        arg = group.add_argument(
            "--label-ids",
            nargs="*",
            default=GoogleMailAPI.default_label_ids(),
            help="match labels",
        )
        self.cli.add_default_to_help(arg, group)

        group.add_argument(
            "--has-attachments",
            action="store_true",
            help="search messages with any files attached",
        )

        group.add_argument(
            "--has-images",
            action="store_true",
            help="search messages with image files attached",
        )

        group.add_argument(
            "--has-videos",
            action="store_true",
            help="search messages with video files attached",
        )

        group.add_argument(
            "--search-query",
            help="gmail search box query pattern",
        )

        self.add_limit_option(parser)

    def run(self) -> None:
        """Run mail `list` command."""

        if self.options.msg_id:
            msg = self.cli.api.get_message(self.options.msg_id)
            self.display_message(msg)
            return

        # See https://support.google.com/mail/answer/7190?hl=en

        if self.options.has_attachments:
            self.options.search_query = "has:attachment"

        elif self.options.has_images:
            self.options.search_query = "filename:(jpg OR jpeg OR png OR tiff OR bmp OR pdf)"

        elif self.options.has_videos:
            self.options.search_query = "filename:(mp4 OR wmv OR mov OR mpg)"

        print(
            str.format(
                "Searching for {!r} in {!r}", self.options.search_query, self.options.label_ids
            )
        )

        for idx, msg_id in enumerate(
            self.cli.api.get_next_msg_id(
                label_ids=self.options.label_ids,
                search_query=self.options.search_query,
            )
        ):
            if self.check_limit():
                break

            if not self.options.print_listing:
                logger.info("Message {} id {!r}", idx + 1, msg_id)

            if (
                self.options.pretty_print
                or self.options.print_listing
                or self.options.print_message
            ):
                msg = self.cli.api.get_message(msg_id)
                self.display_message(msg)

    def display_message(self, msg: dict[str, Any]) -> None:
        """Display `msg`."""

        if self.options.pretty_print:
            self.pprint(msg, max_string=200)
        elif self.options.print_listing:
            self.print_listing(msg)
        elif self.options.print_message:
            self.print_message(msg)

    def print_message(self, msg: dict[str, Any]) -> None:
        """docstring."""

        # logger.debug("msg {!r}", msg)

        for key, value in msg.items():
            self._print_item("MSG", key, value)

        self._print_item(
            "MSG",
            "LOCALTIME",
            strftime("%Y-%m-%d %H:%M:%S %Z", localtime(int(msg["internalDate"]) / 1000)),
        )

        #
        payload = msg["payload"]

        for key, value in payload.items():
            self._print_item("PAYLOAD", key, value)

        #
        headers = payload["headers"]
        for hdr_dict in headers:
            key = hdr_dict["name"]
            value = hdr_dict["value"]
            self._print_item("HEADER", key, value)

        try:
            msg_subject = [h["value"] for h in headers if h["name"] == "Subject"][0]
        except IndexError:  # pragma: no cover
            msg_subject = ""  # pragma: no cover
        self._print_item("HEADER", "SUBJECT", msg_subject)

        try:
            msg_from = [h["value"] for h in headers if h["name"] == "From"][0]
        except IndexError:  # pragma: no cover
            msg_from = ""  # pragma: no cover
        self._print_item("HEADER", "FROM", msg_from)

    def print_listing(self, msg: dict[str, Any]) -> None:
        """docstring."""

        # logger.info('msg {!r}', msg.keys())

        # self._print_item('MSG', 'id', msg['id'])
        # self._print_item('MSG', 'snippet', msg['snippet'])
        # self._print_item('MSG', 'sizeEstimate', msg['sizeEstimate'])
        # self._print_item('MSG', 'labelIds', msg['labelIds'])

        timestamp = strftime("%Y-%m-%d %H:%M:%S %Z", localtime(int(msg["internalDate"]) / 1000))

        payload = msg["payload"]
        headers = payload["headers"]

        try:
            msg_subject = [h["value"] for h in headers if h["name"] == "Subject"][0]
        except IndexError:  # pragma: no cover
            msg_subject = ""  # pragma: no cover
        # self._print_item('HEADER', 'SUBJECT', msg_subject)

        try:
            msg_from = [h["value"] for h in headers if h["name"] == "From"][0]
        except IndexError:  # pragma: no cover
            msg_from = ""  # pragma: no cover
        # self._print_item('HEADER', 'FROM', msg_from)

        print(str.format("{} {:<40} {}", timestamp, msg_from, msg_subject))

    @staticmethod
    def _print_item(tag: str, key: str, value: str) -> None:
        """docstring."""
        truncated = str(value)[:90] + (str(value)[90:] and "...")
        print(str.format("{!r:<7} {!r:<20} {!r}", tag, key, truncated))



import os
import sys
from argparse import Namespace

import pytest

from gmail.api import GoogleMailAPI
from gmail.cli import GoogleMailCLI

# import sys
# from loguru import logger
# logger.add(sys.stdout, level="TRACE")

slow = pytest.mark.skipif(not os.environ.get("SLOW"), reason="slow")


@pytest.fixture(name="mail")
def mail_() -> GoogleMailAPI:
    options = Namespace()
    return GoogleMailAPI(options)


def run_cli(options: list[str]) -> None:
    """Test calling the cli directly."""

    sys.argv = ["gmail"]
    if options:
        sys.argv += options
    print(f"\nRunning {sys.argv!r}", flush=True)
    GoogleMailCLI().main()


def test_gmail_no_args() -> None:
    with pytest.raises(SystemExit) as err:
        run_cli([])
    assert err.value.code == 2


def test_gmail_labels() -> None:
    run_cli(["labels"])


def test_gmail_labels_pretty_print() -> None:
    run_cli(["labels", "--pretty"])


def test_gmail_labels_limit_3() -> None:
    run_cli(["labels", "--limit", "3"])


@slow
def test_gmail_labels_show_counts() -> None:
    run_cli(["labels", "--show-counts"])


def test_gmail_list_limit_20() -> None:
    run_cli(["list", "--limit", "20"])


def test_list_pretty_print_limit_1() -> None:
    run_cli(["list", "--pretty-print", "--limit", "1"])


def test_list_print_message_limit_1() -> None:
    run_cli(["list", "--print-message", "--limit", "1"])


def test_list_print_message_msg_id(mail: GoogleMailAPI) -> None:

    for i, msg_id in enumerate(mail.get_next_msg_id()):
        run_cli(["list", "--print-message", "--msg-id", msg_id])
        if i >= 2:
            break


def test_list_print_listing_limit_3() -> None:
    run_cli(["list", "--print-listing", "--limit", "3"])


def test_list_has_attachments_limit_3() -> None:
    run_cli(["list", "--has-attachments", "--print-listing", "--limit", "3"])


def test_list_has_images_limit_3() -> None:
    run_cli(["list", "--has-images", "--print-listing", "--limit", "3"])


def test_list_has_videos_limit_3() -> None:
    run_cli(["list", "--has-videos", "--print-listing", "--limit", "3"])


@slow
def _test_get_next_msg_id_label_ids_1(mail: GoogleMailAPI) -> None:
    label_ids = ["INBOX"]
    count = sum(1 for _ in mail.get_next_msg_id(label_ids=label_ids))
    print(f"labels_ids {label_ids!r} count(messages) {count}")


def test_get_next_msg_id_label_ids_2(mail: GoogleMailAPI) -> None:
    label_ids = ["INBOX", "UNREAD"]
    count = sum(1 for _ in mail.get_next_msg_id(label_ids=label_ids))
    print(f"labels_ids {label_ids!r} count(messages) {count}")


def test_get_next_msg_id_search_query_1(mail: GoogleMailAPI) -> None:
    search_query = "filename:pdf"
    count = sum(1 for _ in mail.get_next_msg_id(search_query=search_query))
    print(f"search_query {search_query!r} count(messages) {count}")


def test_get_next_msg_id_search_query_2(mail: GoogleMailAPI) -> None:
    search_query = "has:attachment"
    count = sum(1 for _ in mail.get_next_msg_id(search_query=search_query))
    print(f"search_query {search_query!r} count(messages) {count}")


def test_get_next_attachment_id_1(mail: GoogleMailAPI) -> None:
    search_query = "has:attachment"
    for i, msg_id in enumerate(mail.get_next_msg_id(search_query=search_query)):
        for attachment_id in mail.get_next_attachment_id(msg_id):
            print(f"msg_id {msg_id!r} attachment_id {attachment_id!r}")
        if i >= 2:
            break


def test_download_msg_id(mail: GoogleMailAPI) -> None:
    search_query = "has:attachment"
    for i, msg_id in enumerate(mail.get_next_msg_id(search_query=search_query)):
        run_cli(["download", msg_id])
        if i >= 2:
            break



"""Mail `labels` command module."""

from loguru import logger

from gmail.commands import GoogleMailCmd


class MailLabelsCmd(GoogleMailCmd):
    """Mail `labels` command class."""

    def init_command(self) -> None:
        """Initialize mail `labels` command."""

        parser = self.add_subcommand_parser(
            "labels",
            help="list labels",
            description=self.cli.dedent(
                """
    The `%(prog)s` program lists labels.
                """
            ),
        )

        parser.add_argument(
            "--show-counts",
            action="store_true",
            help="show message counts",
        )

        self.add_limit_option(parser)
        self.add_pretty_print_option(parser)

    def run(self) -> None:
        """Run mail `labels` command."""

        labels = self.cli.api.get_labels()
        nlabels = len(labels)

        print("There are", nlabels, "labels")

        for idx, label in enumerate(labels):

            if self.check_limit():
                break

            if self.cli.options.pretty_print:
                self.pprint(label)
                continue

            logger.debug("idx {} label {!r}", idx, label)

            label_id = label["id"]
            label_name = label["name"]

            if self.options.show_counts:
                n_msg_ids = sum(1 for _ in self.cli.api.get_next_msg_id(label_ids=[label_id]))

                print(
                    str.format(
                        "Label {:3} of {:3}: {:7} msgs id {:20} name {}",
                        idx + 1,
                        nlabels,
                        n_msg_ids,
                        label_id,
                        label_name,
                    )
                )
            else:
                print(
                    str.format(
                        "Label {:3} of {:3}: id {:20} name {}",
                        idx + 1,
                        nlabels,
                        label_id,
                        label_name,
                    )
                )



"""Google Mail Commands."""

from argparse import ArgumentParser, _ArgumentGroup
from typing import Any, TypeVar

from libcli import BaseCmd
from loguru import logger
from rich.pretty import pprint as rich_pretty_print

from gmail.cli import GoogleMailCLI

Parser = TypeVar("Parser", ArgumentParser, _ArgumentGroup)

__all__ = ["GoogleMailCmd"]


class GoogleMailCmd(BaseCmd):
    """Base class of google mail commands."""

    cli: GoogleMailCLI

    def add_limit_option(self, parser: Parser) -> None:
        """Add `--limit` to the given `parser`."""

        parser.add_argument(
            "--limit",
            type=int,
            help="limit execution to `LIMIT` number of items",
        )

    def check_limit(self) -> bool:
        """Call at top of loop before performing work."""

        if self.options.limit is None:
            logger.trace("No limit")
            return False

        self.options.limit -= 1
        logger.trace("limit {!r}", self.options.limit)
        assert isinstance(self.options.limit, int)
        return self.options.limit < 0

    def add_pretty_print_option(self, parser: Parser) -> None:
        """Add `--pretty-print` to the given `parser`."""

        parser.add_argument(
            "--pretty-print",
            action="store_true",
            help="pretty-print items",
        )

    @staticmethod
    def pprint(obj: Any, **kwargs: Any) -> None:
        """Make `pprint` convenient."""

        rich_pretty_print(obj, **kwargs)



"""Command Line Interface to Google Mail."""

from pathlib import Path

from libcli import BaseCLI

from gmail.api import GoogleMailAPI

__all__ = ["GoogleMailCLI"]


class GoogleMailCLI(BaseCLI):
    """Command Line Interface to Google Mail."""

    config = {
        # name of config file.
        "config-file": Path("~/.libgoogle.toml"),
        # toml [section-name].
        "config-name": "gmail",
        # distribution name, not importable package name
        "dist-name": "rlane-gmail",
    }

    api: GoogleMailAPI  # connection to google service

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.ArgumentParser(
            prog=__package__,
            description="Google `mail` command line interface.",
            epilog="See `%(prog)s COMMAND --help` for help on a specific command.",
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        self.add_subcommand_modules("gmail.commands", prefix="Mail", suffix="Cmd")

    def main(self) -> None:
        """Command line interface entry point (method)."""

        if not self.options.cmd:
            self.parser.print_help()
            self.parser.exit(2, "error: Missing COMMAND\n")

        self.api = GoogleMailAPI(self.options)
        self.options.cmd()


def main(args: list[str] | None = None) -> None:
    """Command line interface entry point (function)."""
    GoogleMailCLI(args).main()



"""Interface to Google Mail."""

import base64
import os
from argparse import Namespace
from typing import Any, Iterator

import libgoogle
import xdg
from loguru import logger

__all__ = ["GoogleMailAPI"]


class GoogleMailAPI:
    """Interface to Google Mail.

    See https://developers.google.com/gmail/api/v1/reference
    """

    mimetype_PDF = "application/pdf"

    def __init__(self, options: Namespace) -> None:
        """Connect to Google Mail."""

        self.options = options
        self.service = libgoogle.connect("gmail.readonly", "v1")

        self.download_dir = xdg.xdg_data_home() / "gmail"
        self.user_id = "me"

    @staticmethod
    def default_label_ids() -> list[str]:
        """Return list of default label ids."""
        return ["INBOX"]

    def get_labels(self) -> list[dict[str, str]]:
        """Return all labels in the user's mailbox."""

        # https://developers.google.com/gmail/api/v1/reference/users/labels/list

        parms = {"userId": self.user_id}

        logger.debug("service.users().labels().list({!r})", parms)
        response: dict[str, list[Any]] = (
            self.service.users().labels().list(**parms).execute()  # noqa: PLE101
        )  # noqa: PLE101
        logger.trace("response {!r}", response)

        assert isinstance(response, dict)

        if not (labels := response.get("labels")):
            return []  # pragma: no cover

        assert isinstance(labels, list)
        return labels

    def get_next_msg_id(
        self,
        label_ids: list[str] | None = None,
        search_query: str | None = None,
    ) -> Iterator[str]:
        """Return the messages in the user's mailbox."""

        # https://developers.google.com/gmail/api/v1/reference/users/messages/list

        parms = {
            "userId": self.user_id,
            "labelIds": label_ids,
            "q": search_query,
        }

        while True:
            logger.debug("service.users().messages().list({!r})", parms)
            response = self.service.users().messages().list(**parms).execute()  # noqa: PLE101
            logger.trace("response {!r}", response)

            for _ in response.get("messages", []):
                yield _["id"]

            parms["pageToken"] = response.get("nextPageToken")
            if parms["pageToken"] is None:
                return

    def get_message(self, msg_id: str) -> dict[str, Any]:
        """Return specified message."""

        # https://developers.google.com/gmail/api/v1/reference/users/messages/get

        parms = {
            "userId": self.user_id,
            "id": msg_id,
        }

        logger.debug("service.users().messages().get({!r})", parms)
        response = self.service.users().messages().get(**parms).execute()  # noqa: PLE101
        logger.trace("response {!r}", response)

        assert isinstance(response, dict)
        return response

    def get_next_attachment_id(self, msg_id: str) -> Iterator[tuple[str, str, str]]:
        """docstring."""

        msg = self.get_message(msg_id)

        payload = msg.get("payload")
        if not payload:
            logger.debug("No payload")  # pragma: no cover
            return  # pragma: no cover

        parts = payload.get("parts", [])
        if not parts:
            logger.debug("No parts")  # pragma: no cover
            return  # pragma: no cover

        for part in self._flatten_nested_email_parts(parts):

            mimetype = part.get("mimeType")

            (basename, ext) = os.path.splitext(part["filename"])
            filename = os.path.join(self.download_dir, basename + "-" + msg_id + ext)

            body: dict[str, str] | None = part.get("body")
            attachment_id = body.get("attachmentId") if body else None

            logger.debug(
                "Part mimeType {!r} filename {!r} attachment_id {!r}",
                mimetype,
                filename,
                attachment_id,
            )

            if not mimetype:
                logger.debug("Missing mimeType")  # pragma: no cover
                continue  # pragma: no cover
            if not filename:
                logger.debug("Missing filename")  # pragma: no cover
                continue  # pragma: no cover
            if not body:
                logger.debug("Missing body")  # pragma: no cover
                continue  # pragma: no cover
            if not attachment_id:
                logger.debug("Missing attachmentId")
                continue

            yield mimetype, filename, attachment_id

    def get_attachment_data(self, msg_id: str, attachment_id: str) -> bytes:
        """docstring."""

        att = (
            self.service.users()  # noqa: PLE101
            .messages()
            .attachments()
            .get(userId=self.user_id, id=attachment_id, messageId=msg_id)
            .execute()
        )

        # LANG=en_US.UTF-8
        data = att["data"]
        data = data.replace("-", "+")
        data = data.replace("_", "/")
        data = base64.b64decode(bytes(data, "UTF-8"))

        return bytes(data)

    @staticmethod
    def _flatten_nested_email_parts(parts: list[dict[str, Any]]) -> list[dict[str, Any]]:
        all_parts: list[dict[str, Any]] = []
        for part in parts:
            if p := part.get("parts"):
                all_parts.extend(p)
            else:
                all_parts.append(part)
        return all_parts
